# -*- coding: utf-8 -*-
'''
Tests to try out salt key.RaetKey Potentially ephemeral
'''
from __future__ import absolute_import, print_function, unicode_literals
import sys
from salt.ext.six.moves import map
# pylint: disable=blacklisted-import
if sys.version_info < (2, 7):
    import unittest2 as unittest
else:
    import unittest
# pylint: enable=blacklisted-import

import os
import shutil
import socket
import stat
import tempfile
import time

from ioflo.aid.odicting import odict
from ioflo.aid.timing import StoreTimer
from ioflo.base import storing
from ioflo.base.consoling import getConsole
console = getConsole()

from raet import raeting, nacling
from raet.road import estating, stacking

from salt.daemons import salting
import salt.utils.kinds as kinds
import salt.utils.stringutils


def setUpModule():
    console.reinit(verbosity=console.Wordage.concise)


def tearDownModule():
    pass


class BasicTestCase(unittest.TestCase):  # pylint: disable=moved-test-case-class

    def setUp(self):
        self.store = storing.Store(stamp=0.0)
        self.timer = StoreTimer(store=self.store, duration=1.0)

        self.tempDirpath = tempfile.mkdtemp(prefix="salt", suffix='keep', dir='/tmp')
        self.maxDiff = None

    def tearDown(self):
        if os.path.exists(self.tempDirpath):
            shutil.rmtree(self.tempDirpath)

    def createOpts(self,
                   role,
                   kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                   dirpath='/tmp',
                   openMode=False,
                   autoAccept=True):
        '''
        Create associated pki directories for stack and return opts

        os.path.join(cache, 'raet', name, 'remote')
        '''
        pkiDirpath = os.path.join(dirpath, 'pki', role, 'raet')
        if not os.path.exists(pkiDirpath):
            os.makedirs(pkiDirpath)

        acceptedDirpath = os.path.join(pkiDirpath, 'accepted')
        if not os.path.exists(acceptedDirpath):
            os.makedirs(acceptedDirpath)

        pendingDirpath = os.path.join(pkiDirpath, 'pending')
        if not os.path.exists(pendingDirpath):
            os.makedirs(pendingDirpath)

        rejectedDirpath = os.path.join(pkiDirpath, 'rejected')
        if not os.path.exists(rejectedDirpath):
            os.makedirs(rejectedDirpath)

        localFilepath = os.path.join(pkiDirpath, 'local.key')
        if os.path.exists(localFilepath):
            mode = os.stat(localFilepath).st_mode
            print(mode)
            os.chmod(localFilepath, mode | stat.S_IWUSR | stat.S_IWUSR)

        cacheDirpath = os.path.join(dirpath, 'cache', role)
        sockDirpath = os.path.join(dirpath, 'sock', role)

        opts = dict(
                     id=role,
                     pki_dir=pkiDirpath,
                     sock_dir=sockDirpath,
                     cachedir=cacheDirpath,
                     open_mode=openMode,
                     auto_accept=autoAccept,
                     transport='raet',
                     __role=kind,
                     )
        return opts

    def createRoadData(self, role, kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master], cachedirpath=''):
        '''
        Creates odict and populates with data to setup road stack
        {
            name: stack name local estate name
            dirpath: dirpath for keep files
            sighex: signing key
            verhex: verify key
            prihex: private key
            pubhex: public key
        }
        '''
        data = odict()
        data['name'] = "{0}_{1}".format(role, kind)
        data['role'] = role
        data['kind'] = kinds.APPL_KINDS[kind]  # convert to integer from kind name
        data['basedirpath'] = os.path.join(cachedirpath, 'raet')
        signer = nacling.Signer()
        data['sighex'] = signer.keyhex
        data['verhex'] = signer.verhex
        privateer = nacling.Privateer()
        data['prihex'] = privateer.keyhex
        data['pubhex'] = privateer.pubhex

        return data

    def createRoadStack(self, data, keep, uid=None, main=None, ha=None, mutable=None):
        '''
        Creates stack and local estate from data with
        local estate.uid = uid
        stack.main = main
        stack.mutable = mutable
        stack.auto = auto
        stack.name = data['name']
        local estate.name = data['name']
        local estate.ha = ha

        returns stack

        '''
        stack = stacking.RoadStack(store=self.store,
                                   name=data['name'],
                                   keep=keep,
                                   uid=uid,
                                   ha=ha,
                                   main=main,
                                   mutable=mutable,
                                   role=data['role'],
                                   kind=data['kind'],
                                   sigkey=data['sighex'],
                                   prikey=data['prihex'],)

        return stack

    def join(self, initiator, correspondent, deid=None, duration=1.0):
        '''
        Utility method to do join. Call from test method.
        '''
        console.terse("\nJoin Transaction **************\n")
        if not initiator.remotes:
            remote = initiator.addRemote(estating.RemoteEstate(stack=initiator,
                                                      fuid=0,  # vacuous join
                                                      sid=0,  # always 0 for join
                                                      ha=correspondent.local.ha))
            deid = remote.uid
        initiator.join(uid=deid)
        self.service(correspondent, initiator, duration=duration)

    def allow(self, other, main, duration=1.0):
        '''
        Utility method to do allow. Call from test method.
        '''
        console.terse("\nAllow Transaction **************\n")
        other.allow()
        self.service(main, other, duration=duration)

    def message(self, main, other, mains, others, duration=2.0):
        '''
        Utility to send messages both ways
        '''
        for msg in mains:
            main.transmit(msg)
        for msg in others:
            other.transmit(msg)

        self.service(main, other, duration=duration)

    def service(self, main, other, duration=1.0):
        '''
        Utility method to service queues. Call from test method.
        '''
        self.timer.restart(duration=duration)
        while not self.timer.expired:
            other.serviceAll()
            main.serviceAll()
            if not (main.transactions or other.transactions):
                break
            self.store.advanceStamp(0.05)
            time.sleep(0.05)

    def testBasic(self):
        '''
        Basic keep setup for stack keep persistence load and dump in normal mode
        '''
        console.terse("{0}\n".format(self.testBasic.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=False)
        mainData = self.createRoadData(cachedirpath=opts['cachedir'],
                                       role=opts['id'],
                                       kind=opts['__role'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(
                os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.never.value)
        self.assertDictEqual(main.keep.loadLocalData(),
                             {'name': mainData['name'],
                              'uid': 1,
                              'ha': ['127.0.0.1', 7530],
                              'iha': None,
                              'natted': None,
                              'fqdn': socket.getfqdn('127.0.0.1'),
                              'dyned': None,
                              'sid': 0,
                              'puid': 1,
                              'aha': ['0.0.0.0', 7530],
                              'role': mainData['role'],
                              'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                              'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                              })

        data1 = self.createRoadData(role='remote1',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data1['name'],
                                             kind=data1['kind'],
                                             ha=('127.0.0.1', 7532),
                                             role=data1['role'],
                                             verkey=data1['verhex'],
                                             pubkey=data1['pubhex'],))

        data2 = self.createRoadData(role='remote2',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data2['name'],
                                             kind=data2['kind'],
                                             ha=('127.0.0.1', 7533),
                                             role=data2['role'],
                                             verkey=data2['verhex'],
                                             pubkey=data2['pubhex'],))

        main.dumpRemotes()

        # acceptance will be pended

        self.assertDictEqual(main.keep.loadAllRemoteData(),
            {
                'remote1_minion':
                    {'name': data1['name'],
                     'uid': 2,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7532],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data1['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 0,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     },
                'remote2_minion':
                    {'name': data2['name'],
                     'uid': 3,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7533],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data2['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data2['role'],
                     'acceptance': 0,
                     'verhex': salt.utils.stringutils.to_str(data2['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data2['pubhex']),
                     }
            })

        # now recreate with saved data
        main.server.close()
        mainKeep = salting.SaltKeep(opts=opts,
                                     basedirpath=mainData['basedirpath'],
                                     stackname=mainData['name'])
        main = stacking.RoadStack(name=mainData['name'],
                                  store=self.store,
                                  keep=mainKeep,
                                  main=True)

        self.assertEqual(main.local.priver.keyhex, mainData['prihex'])
        self.assertEqual(main.local.signer.keyhex, mainData['sighex'])

        self.assertEqual(len(main.remotes), 2)

        # other stack
        opts = self.createOpts(role='other',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=False)
        otherData = self.createRoadData(role=opts['id'],
                                        kind=opts['__role'],
                                        cachedirpath=opts['cachedir'])
        otherKeep = salting.SaltKeep(opts=opts,
                                      basedirpath=otherData['basedirpath'],
                                      stackname=otherData['name'])

        self.assertEqual(otherKeep.loadLocalData(), None)
        self.assertEqual(otherKeep.loadAllRemoteData(), {})

        other = self.createRoadStack(data=otherData,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=otherKeep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other.name, other.keep.dirpath))
        self.assertTrue(other.keep.dirpath.endswith(os.path.join('other', 'raet', 'other_minion')))
        self.assertEqual(other.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other.keep.auto, raeting.AutoMode.never.value)

        self.assertDictEqual(other.keep.loadLocalData(),
                            {
                                'name': otherData['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': otherData['role'],
                                'sighex': salt.utils.stringutils.to_str(otherData['sighex']),
                                'prihex': salt.utils.stringutils.to_str(otherData['prihex']),
                            })

        data3 = self.createRoadData(role='remote3',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        other.addRemote(estating.RemoteEstate(stack=other,
                                              name=data3['name'],
                                              kind=data3['kind'],
                                              ha=('127.0.0.1', 7534),
                                              role=data3['role'],
                                              verkey=data3['verhex'],
                                              pubkey=data3['pubhex'],))

        data4 = self.createRoadData(role='remote4',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        other.addRemote(estating.RemoteEstate(stack=other,
                                              name=data4['name'],
                                              kind=data4['kind'],
                                              ha=('127.0.0.1', 7535),
                                              role=data4['role'],
                                              verkey=data4['verhex'],
                                              pubkey=data4['pubhex'],))

        other.dumpRemotes()
        self.assertDictEqual(other.keep.loadAllRemoteData(),
            {
                'remote3_minion':
                {
                    'name': data3['name'],
                    'uid': 2,
                    'fuid': 0,
                    'ha': ['127.0.0.1', 7534],
                    'iha': None,
                    'natted': None,
                    'fqdn': socket.getfqdn('127.0.0.1'),
                    'dyned': None,
                    'main': False,
                    'kind': data3['kind'],
                    'sid': 0,
                    'joined': None,
                    'role': data3['role'],
                    'acceptance': 0,
                    'verhex': salt.utils.stringutils.to_str(data3['verhex']),
                    'pubhex': salt.utils.stringutils.to_str(data3['pubhex']),
                },
                'remote4_minion':
                {
                    'name': data4['name'],
                    'uid': 3,
                    'fuid': 0,
                    'ha': ['127.0.0.1', 7535],
                    'iha': None,
                    'natted': None,
                    'fqdn': socket.getfqdn('127.0.0.1'),
                    'dyned': None,
                    'main': False,
                    'kind': data4['kind'],
                    'sid': 0,
                    'joined': None,
                    'role': data4['role'],
                    'acceptance': 0,
                    'verhex': salt.utils.stringutils.to_str(data4['verhex']),
                    'pubhex': salt.utils.stringutils.to_str(data4['pubhex']),
                }
            })

        main.server.close()
        other.server.close()

    def testBasicOpen(self):
        '''
        Basic keep setup for stack keep persistence load and dump in open mode
        '''
        console.terse("{0}\n".format(self.testBasicOpen.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=True,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.always.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        data1 = self.createRoadData(role='remote1',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data1['name'],
                                             kind=data1['kind'],
                                             ha=('127.0.0.1', 7532),
                                             role=data1['role'],
                                             verkey=data1['verhex'],
                                             pubkey=data1['pubhex'],))

        data2 = self.createRoadData(role='remote2',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data2['name'],
                                             kind=data2['kind'],
                                             ha=('127.0.0.1', 7533),
                                             role=data2['role'],
                                             verkey=data2['verhex'],
                                             pubkey=data2['pubhex'],))

        main.dumpRemotes()

        self.assertDictEqual(main.keep.loadAllRemoteData(),
            {
                'remote1_minion':
                    {'name': data1['name'],
                     'uid': 2,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7532],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data1['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     },
                'remote2_minion':
                    {'name': data2['name'],
                     'uid': 3,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7533],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data2['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data2['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data2['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data2['pubhex']),
                     }
            })

        # now recreate with saved data
        main.server.close()
        mainKeep = salting.SaltKeep(opts=opts,
                                     basedirpath=mainData['basedirpath'],
                                     stackname=mainData['name'])
        main = stacking.RoadStack(name=mainData['name'],
                                  store=self.store,
                                  keep=mainKeep,
                                  main=True)

        self.assertEqual(main.local.priver.keyhex, mainData['prihex'])
        self.assertEqual(main.local.signer.keyhex, mainData['sighex'])

        self.assertEqual(len(main.remotes), 2)

        # other stack
        opts = self.createOpts(role='other',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=True,
                               autoAccept=True)
        otherData = self.createRoadData(role='other',
                                        kind=opts['__role'],
                                        cachedirpath=opts['cachedir'])
        otherKeep = salting.SaltKeep(opts=opts,
                                      basedirpath=otherData['basedirpath'],
                                      stackname=otherData['name'])

        self.assertEqual(otherKeep.loadLocalData(), None)
        self.assertEqual(otherKeep.loadAllRemoteData(), {})

        other = self.createRoadStack(data=otherData,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=otherKeep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other.name, other.keep.dirpath))
        self.assertTrue(other.keep.dirpath.endswith(os.path.join('other', 'raet', 'other_minion')))
        self.assertEqual(other.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other.keep.auto, raeting.AutoMode.always.value)

        self.assertDictEqual(other.keep.loadLocalData(),
                            {
                                'name': otherData['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': otherData['role'],
                                'sighex': salt.utils.stringutils.to_str(otherData['sighex']),
                                'prihex': salt.utils.stringutils.to_str(otherData['prihex']),
                            })

        data3 = self.createRoadData(role='remote3',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        other.addRemote(estating.RemoteEstate(stack=other,
                                              name=data3['name'],
                                              kind=data3['kind'],
                                              ha=('127.0.0.1', 7534),
                                              role=data3['role'],
                                              verkey=data3['verhex'],
                                              pubkey=data3['pubhex'],))

        data4 = self.createRoadData(role='remote4',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        other.addRemote(estating.RemoteEstate(stack=other,
                                              name=data4['name'],
                                              kind=data4['kind'],
                                              ha=('127.0.0.1', 7535),
                                              role=data4['role'],
                                              verkey=data4['verhex'],
                                              pubkey=data4['pubhex'],))

        other.dumpRemotes()
        self.assertDictEqual(other.keep.loadAllRemoteData(),
            {
                'remote3_minion':
                {
                    'name': data3['name'],
                    'uid': 2,
                    'fuid': 0,
                    'ha': ['127.0.0.1', 7534],
                    'iha': None,
                    'natted': None,
                    'fqdn': socket.getfqdn('127.0.0.1'),
                    'dyned': None,
                    'main': False,
                    'kind': data3['kind'],
                    'sid': 0,
                    'joined': None,
                    'role': data3['role'],
                    'acceptance': 1,
                    'verhex': salt.utils.stringutils.to_str(data3['verhex']),
                    'pubhex': salt.utils.stringutils.to_str(data3['pubhex']),
                },
                'remote4_minion':
                {
                    'name': data4['name'],
                    'uid': 3,
                    'fuid': 0,
                    'ha': ['127.0.0.1', 7535],
                    'iha': None,
                    'natted': None,
                    'fqdn': socket.getfqdn('127.0.0.1'),
                    'dyned': None,
                    'main': False,
                    'kind': data4['kind'],
                    'sid': 0,
                    'joined': None,
                    'role': data4['role'],
                    'acceptance': 1,
                    'verhex': salt.utils.stringutils.to_str(data4['verhex']),
                    'pubhex': salt.utils.stringutils.to_str(data4['pubhex']),
                }
            })

        main.server.close()
        other.server.close()

    def testBasicAuto(self):
        '''
        Basic keep setup for stack keep persistence load and dump with auto accept
        '''
        console.terse("{0}\n".format(self.testBasicAuto.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         'role': mainData['role'],
                                                         })

        data1 = self.createRoadData(role='remote1',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data1['name'],
                                             kind=data1['kind'],
                                             ha=('127.0.0.1', 7532),
                                             role=data1['role'],
                                             verkey=data1['verhex'],
                                             pubkey=data1['pubhex'],))

        data2 = self.createRoadData(role='remote2',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data2['name'],
                                             kind=data2['kind'],
                                             ha=('127.0.0.1', 7533),
                                             role=data2['role'],
                                             verkey=data2['verhex'],
                                             pubkey=data2['pubhex'],))

        main.dumpRemotes()

        self.assertDictEqual(main.keep.loadAllRemoteData(),
            {
                'remote1_minion':
                    {
                     'name': data1['name'],
                     'uid': 2,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7532],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data1['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     },
                'remote2_minion':
                    {
                     'name': data2['name'],
                     'uid': 3,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7533],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data2['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data2['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data2['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data2['pubhex']),
                     }
            })

        # now recreate with saved data
        main.server.close()
        mainKeep = salting.SaltKeep(opts=opts,
                                     basedirpath=mainData['basedirpath'],
                                     stackname=mainData['name'])
        main = stacking.RoadStack(name=mainData['name'],
                                  store=self.store,
                                  keep=mainKeep,
                                  main=True)

        self.assertEqual(main.local.priver.keyhex, mainData['prihex'])
        self.assertEqual(main.local.signer.keyhex, mainData['sighex'])

        self.assertEqual(len(main.remotes), 2)

        # other stack
        opts = self.createOpts(role='other',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        otherData = self.createRoadData(role='other',
                                        kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                        cachedirpath=opts['cachedir'])
        otherKeep = salting.SaltKeep(opts=opts,
                                      basedirpath=otherData['basedirpath'],
                                      stackname=otherData['name'])

        self.assertEqual(otherKeep.loadLocalData(), None)
        self.assertEqual(otherKeep.loadAllRemoteData(), {})

        other = self.createRoadStack(data=otherData,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=otherKeep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other.name, other.keep.dirpath))
        self.assertTrue(other.keep.dirpath.endswith(os.path.join('other', 'raet', 'other_minion')))
        self.assertEqual(other.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other.keep.auto, raeting.AutoMode.once.value)

        self.assertDictEqual(other.keep.loadLocalData(),
                            {
                                'name': otherData['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': otherData['role'],
                                'sighex': salt.utils.stringutils.to_str(otherData['sighex']),
                                'prihex': salt.utils.stringutils.to_str(otherData['prihex']),
                            })

        data3 = self.createRoadData(role='remote3',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        other.addRemote(estating.RemoteEstate(stack=other,
                                              name=data3['name'],
                                              kind=data3['kind'],
                                              ha=('127.0.0.1', 7534),
                                              role=data3['role'],
                                              verkey=data3['verhex'],
                                              pubkey=data3['pubhex'],))

        data4 = self.createRoadData(role='remote4',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'])
        other.addRemote(estating.RemoteEstate(stack=other,
                                              name=data4['name'],
                                              kind=data4['kind'],
                                              ha=('127.0.0.1', 7535),
                                              role=data4['role'],
                                              verkey=data4['verhex'],
                                              pubkey=data4['pubhex'],))

        other.dumpRemotes()
        self.assertDictEqual(other.keep.loadAllRemoteData(),
            {
                'remote3_minion':
                {
                    'name': data3['name'],
                    'uid': 2,
                    'fuid': 0,
                    'ha': ['127.0.0.1', 7534],
                    'iha': None,
                    'natted': None,
                    'fqdn': socket.getfqdn('127.0.0.1'),
                    'dyned': None,
                    'main': False,
                    'kind': data3['kind'],
                    'sid': 0,
                    'joined': None,
                    'role': data3['role'],
                    'acceptance': 1,
                    'verhex': salt.utils.stringutils.to_str(data3['verhex']),
                    'pubhex': salt.utils.stringutils.to_str(data3['pubhex']),
                },
                'remote4_minion':
                {
                    'name': data4['name'],
                    'uid': 3,
                    'fuid': 0,
                    'ha': ['127.0.0.1', 7535],
                    'iha': None,
                    'natted': None,
                    'fqdn': socket.getfqdn('127.0.0.1'),
                    'dyned': None,
                    'main': False,
                    'kind': data4['kind'],
                    'sid': 0,
                    'joined': None,
                    'role': data4['role'],
                    'acceptance': 1,
                    'verhex': salt.utils.stringutils.to_str(data4['verhex']),
                    'pubhex': salt.utils.stringutils.to_str(data4['pubhex']),
                }
            })

        main.server.close()
        other.server.close()

    def testBasicRole(self):
        '''
        Basic keep setup for stack keep persistence load and dump with shared role
        '''
        console.terse("{0}\n".format(self.testBasicRole.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=False)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'],
                                       )
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.never.value)
        self.assertDictEqual(main.keep.loadLocalData(), {'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        # add multiple remotes all with same role
        data1 = self.createRoadData(role='primary',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'],
                                    )
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data1['name'],
                                             kind=data1['kind'],
                                             ha=('127.0.0.1', 7532),
                                             role=data1['role'],
                                             verkey=data1['verhex'],
                                             pubkey=data1['pubhex'],
                                             ))

        data2 = self.createRoadData(role='primary',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.caller],
                                    cachedirpath=opts['cachedir'],
                                    )
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data2['name'],
                                             kind=data2['kind'],
                                             ha=('127.0.0.1', 7533),
                                             role=data2['role'],
                                             verkey=data2['verhex'],
                                             pubkey=data2['pubhex'],
                                             ))

        main.dumpRemotes()

        # will save remote1 keys and reuse for remote2

        self.assertDictEqual(main.keep.loadAllRemoteData(),
            {
                'primary_minion':
                    {
                     'name': data1['name'],
                     'uid': 2,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7532],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data1['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 0,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     },
                'primary_caller':
                    {
                     'name': data2['name'],
                     'uid': 3,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7533],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data2['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 0,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     }
            })

        # now recreate with saved data
        main.server.close()
        mainKeep = salting.SaltKeep(opts=opts,
                                     basedirpath=mainData['basedirpath'],
                                     stackname=mainData['name'])
        main = stacking.RoadStack(name=mainData['name'],
                                  store=self.store,
                                  keep=mainKeep,
                                  main=True)

        self.assertEqual(main.local.name, mainData['name'])
        self.assertEqual(main.local.uid, 1)
        self.assertEqual(main.main, True)
        self.assertEqual(main.local.role, mainData['role'])
        self.assertIs(main.keep.auto, raeting.AutoMode.never.value)
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertEqual(main.local.priver.keyhex, mainData['prihex'])
        self.assertEqual(main.local.signer.keyhex, mainData['sighex'])

        self.assertEqual(len(main.remotes), 2)
        for data in [data1, data2]:
            remote = main.nameRemotes[data['name']]
            self.assertEqual(remote.name, data['name'])
            self.assertEqual(remote.role, data['role'])
            self.assertEqual(remote.pubber.keyhex, data1['pubhex'])
            self.assertEqual(remote.verfer.keyhex, data1['verhex'])

        main.server.close()

    def testBasicRoleOpen(self):
        '''
        Basic keep setup for stack keep persistence load and dump with shared role
        '''
        console.terse("{0}\n".format(self.testBasicRoleOpen.__doc__))

        opts = self.createOpts(role='main',
                               dirpath=self.tempDirpath,
                               openMode=True,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'],
)
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.always.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        # add multiple remotes all with same role
        data1 = self.createRoadData(role='primary',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'],)
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data1['name'],
                                             kind=data1['kind'],
                                             ha=('127.0.0.1', 7532),
                                             role=data1['role'],
                                             verkey=data1['verhex'],
                                             pubkey=data1['pubhex'],))

        data2 = self.createRoadData(role='primary',
                                    kind='syndic',
                                    cachedirpath=opts['cachedir'],)
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data2['name'],
                                             kind=data2['kind'],
                                             ha=('127.0.0.1', 7533),
                                             role=data2['role'],
                                             verkey=data2['verhex'],
                                             pubkey=data2['pubhex'],))

        main.dumpRemotes()  # second one keys will clobber first one keys

        self.assertDictEqual(main.keep.loadAllRemoteData(),
            {
                'primary_minion':
                    {
                     'name': data1['name'],
                     'uid': 2,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7532],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data1['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data2['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data2['pubhex']),
                     },
                'primary_syndic':
                    {
                     'name': data2['name'],
                     'uid': 3,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7533],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data2['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data2['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data2['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data2['pubhex']),
                     }
            })

        # now recreate with saved data
        main.server.close()
        mainKeep = salting.SaltKeep(opts=opts,
                                     basedirpath=mainData['basedirpath'],
                                     stackname=mainData['name'])
        main = stacking.RoadStack(name=mainData['name'],
                                  store=self.store,
                                  keep=mainKeep,
                                  main=True)

        self.assertEqual(main.local.name, mainData['name'])
        self.assertEqual(main.local.uid, 1)
        self.assertEqual(main.main, True)
        self.assertEqual(main.local.role, mainData['role'])
        self.assertIs(main.keep.auto, raeting.AutoMode.always.value)
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertEqual(main.local.priver.keyhex, mainData['prihex'])
        self.assertEqual(main.local.signer.keyhex, mainData['sighex'])

        self.assertEqual(len(main.remotes), 2)
        for data in [data1, data2]:
            remote = main.nameRemotes[data['name']]
            self.assertEqual(remote.name, data['name'])
            self.assertEqual(remote.role, data['role'])
            self.assertEqual(remote.pubber.keyhex, data2['pubhex'])
            self.assertEqual(remote.verfer.keyhex, data2['verhex'])

        main.server.close()

    def testBasicRoleAuto(self):
        '''
        Basic keep setup for stack keep persistence load and dump with shared role
        '''
        console.terse("{0}\n".format(self.testBasicRoleAuto.__doc__))
        self.maxDiff = None

        opts = self.createOpts(role='main',
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'],
                                       )
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        # add multiple remotes all with same role but different keys
        data1 = self.createRoadData(role='primary',
                                    kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                                    cachedirpath=opts['cachedir'],
                                    )
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data1['name'],
                                             kind=data1['kind'],
                                             ha=('127.0.0.1', 7532),
                                             role=data1['role'],
                                             verkey=data1['verhex'],
                                             pubkey=data1['pubhex'],
                                             ))

        data2 = self.createRoadData(role='primary',
                                    kind='syndic',
                                    cachedirpath=opts['cachedir'],
                                    )
        main.addRemote(estating.RemoteEstate(stack=main,
                                             name=data2['name'],
                                             kind=data2['kind'],
                                             ha=('127.0.0.1', 7533),
                                             role=data2['role'],
                                             verkey=data2['verhex'],
                                             pubkey=data2['pubhex'],
                                             ))

        main.dumpRemotes()

        # remote2 keys will be rejected since keys do not match remote1 but same role
        # upon reloading the keys for remote2 will be those from remote1

        self.assertDictEqual(main.keep.loadAllRemoteData(),
            {
                'primary_minion':
                    {
                     'name': data1['name'],
                     'uid': 2,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7532],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data1['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data1['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     },
                'primary_syndic':
                    {
                     'name': data2['name'],
                     'uid': 3,
                     'fuid': 0,
                     'ha': ['127.0.0.1', 7533],
                     'iha': None,
                     'natted': None,
                     'fqdn': socket.getfqdn('127.0.0.1'),
                     'dyned': None,
                     'main': False,
                     'kind': data2['kind'],
                     'sid': 0,
                     'joined': None,
                     'role': data2['role'],
                     'acceptance': 1,
                     'verhex': salt.utils.stringutils.to_str(data1['verhex']),
                     'pubhex': salt.utils.stringutils.to_str(data1['pubhex']),
                     }
            })

        # now recreate with saved data
        main.server.close()
        mainKeep = salting.SaltKeep(opts=opts,
                                     basedirpath=mainData['basedirpath'],
                                     stackname=mainData['name'])
        main = stacking.RoadStack(name=mainData['name'],
                                  store=self.store,
                                  keep=mainKeep,
                                  main=True)

        self.assertEqual(main.local.name, mainData['name'])
        self.assertEqual(main.local.uid, 1)
        self.assertEqual(main.main, True)
        self.assertEqual(main.local.role, mainData['role'])
        self.assertIs(main.keep.auto, raeting.AutoMode.once.value)
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertEqual(main.local.priver.keyhex, mainData['prihex'])
        self.assertEqual(main.local.signer.keyhex, mainData['sighex'])

        self.assertEqual(len(main.remotes), 2)
        for data in [data1, data2]:
            remote = main.nameRemotes[data['name']]
            self.assertEqual(remote.name, data['name'])
            self.assertEqual(remote.role, data['role'])
            self.assertEqual(remote.acceptance, 1)
            self.assertEqual(remote.pubber.keyhex, data1['pubhex'])
            self.assertEqual(remote.verfer.keyhex, data1['verhex'])

        main.server.close()

    def testBootstrapNever(self):
        '''
        Bootstap to allowed with never mode on main
        '''
        console.terse("{0}\n".format(self.testBootstrapNever.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=False)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.never.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        opts = self.createOpts(role='other',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        otherData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        otherKeep = salting.SaltKeep(opts=opts,
                                      basedirpath=otherData['basedirpath'],
                                      stackname=otherData['name'])

        self.assertEqual(otherKeep.loadLocalData(), None)
        self.assertEqual(otherKeep.loadAllRemoteData(), {})

        other = self.createRoadStack(data=otherData,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=otherKeep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other.name, other.keep.dirpath))
        self.assertTrue(other.keep.dirpath.endswith(os.path.join('other', 'raet', 'other_minion')))
        self.assertEqual(other.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other.keep.loadLocalData(),
                            {
                                'name': otherData['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': otherData['role'],
                                'sighex': salt.utils.stringutils.to_str(otherData['sighex']),
                                'prihex': salt.utils.stringutils.to_str(otherData['prihex']),
                            })

        self.join(other, main)
        self.assertEqual(len(main.transactions), 1)  # pending
        main.keep.acceptRemote(main.nameRemotes[other.local.name])
        self.service(main, other, duration=1.0)

        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other.transactions), 0)
        remote = next(iter(other.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other.transactions), 0)
        remote = next(iter(other.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # now delete a key and see if road keep file is also deleted
        main.keep.saltRaetKey.delete_key(match=other.local.role)
        remote = main.remotes[2]
        path = os.path.join(main.keep.remotedirpath,
                "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))

        for stack in [main, other]:
            stack.server.close()
            stack.clearAllKeeps()

    def testBootstrapOpen(self):
        '''
        Bootstap to allowed with open mode on main
        '''
        console.terse("{0}\n".format(self.testBootstrapOpen.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=True,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.always.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        opts = self.createOpts(role='other',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        otherData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        otherKeep = salting.SaltKeep(opts=opts,
                                      basedirpath=otherData['basedirpath'],
                                      stackname=otherData['name'])

        self.assertEqual(otherKeep.loadLocalData(), None)
        self.assertEqual(otherKeep.loadAllRemoteData(), {})

        other = self.createRoadStack(data=otherData,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=otherKeep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other.name, other.keep.dirpath))
        self.assertTrue(other.keep.dirpath.endswith(os.path.join('other', 'raet', 'other_minion')))
        self.assertEqual(other.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other.keep.loadLocalData(),
                            {
                                'name': otherData['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': otherData['role'],
                                'sighex': salt.utils.stringutils.to_str(otherData['sighex']),
                                'prihex': salt.utils.stringutils.to_str(otherData['prihex']),
                            })

        self.join(other, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other.transactions), 0)
        remote = next(iter(other.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other.transactions), 0)
        remote = next(iter(other.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # now delete a key and see if road keep file is also deleted
        main.keep.saltRaetKey.delete_key(match=other.local.role)
        remote = main.remotes[2]
        path = os.path.join(main.keep.remotedirpath,
                "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))

        for stack in [main, other]:
            stack.server.close()
            stack.clearAllKeeps()

    def testBootstrapAuto(self):
        '''
        Bootstap to allowed with auto accept on main
        '''
        console.terse("{0}\n".format(self.testBootstrapAuto.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertEqual(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        opts = self.createOpts(role='other',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        otherData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        otherKeep = salting.SaltKeep(opts=opts,
                                      basedirpath=otherData['basedirpath'],
                                      stackname=otherData['name'])

        self.assertEqual(otherKeep.loadLocalData(), None)
        self.assertEqual(otherKeep.loadAllRemoteData(), {})

        other = self.createRoadStack(data=otherData,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=otherKeep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other.name, other.keep.dirpath))
        self.assertTrue(other.keep.dirpath.endswith(os.path.join('other', 'raet', 'other_minion')))
        self.assertEqual(other.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other.keep.loadLocalData(),
                            {
                                'name': otherData['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'sighex': salt.utils.stringutils.to_str(otherData['sighex']),
                                'prihex': salt.utils.stringutils.to_str(otherData['prihex']),
                                'role': otherData['role'],
                            })

        self.join(other, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other.transactions), 0)
        remote = next(iter(other.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other.transactions), 0)
        remote = next(iter(other.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # now delete a key and see if road keep file is also deleted
        main.keep.saltRaetKey.delete_key(match=other.local.role)
        remote = main.remotes[2]
        path = os.path.join(main.keep.remotedirpath,
                "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))

        for stack in [main, other]:
            stack.server.close()
            stack.clearAllKeeps()

    def testBootstrapRoleNever(self):
        '''
        Bootstap to allowed with multiple remotes using same role with never main
        '''
        console.terse("{0}\n".format(self.testBootstrapRoleNever.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=False)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.never.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                         'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        opts = self.createOpts(role='primary',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        other1Data = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        other1Keep = salting.SaltKeep(opts=opts,
                                      basedirpath=other1Data['basedirpath'],
                                      stackname=other1Data['name'])

        self.assertEqual(other1Keep.loadLocalData(), None)
        self.assertEqual(other1Keep.loadAllRemoteData(), {})

        other1 = self.createRoadStack(data=other1Data,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=other1Keep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other1.name, other1.keep.dirpath))
        self.assertTrue(other1.keep.dirpath.endswith(os.path.join('primary', 'raet', 'primary_minion')))
        self.assertEqual(other1.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other1.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other1.keep.loadLocalData(),
                            {
                                'name': other1Data['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': other1Data['role'],
                                'sighex': salt.utils.stringutils.to_str(other1Data['sighex']),
                                'prihex': salt.utils.stringutils.to_str(other1Data['prihex']),
                            })

        self.join(other1, main)
        self.assertEqual(len(main.transactions), 1)  # pending
        main.keep.acceptRemote(main.nameRemotes[other1.local.name])
        self.service(main, other1, duration=1.0)

        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other1.transactions), 0)
        remote = next(iter(other1.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other1, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other1.transactions), 0)
        remote = next(iter(other1.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # create other2 stack but use same role but different keys as other1
        opts = self.createOpts(role='primary',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.caller],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        other2Data = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        other2Keep = salting.SaltKeep(opts=opts,
                                      basedirpath=other2Data['basedirpath'],
                                      stackname=other2Data['name'])

        self.assertEqual(other2Keep.loadLocalData(), None)
        self.assertEqual(other2Keep.loadAllRemoteData(), {})

        other2 = self.createRoadStack(data=other2Data,
                                     main=None,
                                     ha=("", 7532),
                                     keep=other2Keep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other2.name, other2.keep.dirpath))
        self.assertTrue(other2.keep.dirpath.endswith(os.path.join('primary', 'raet', 'primary_caller')))
        self.assertEqual(other2.ha, ("0.0.0.0", 7532))
        self.assertIs(other2.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other2.keep.loadLocalData(),
                            {
                                'name': other2Data['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7532],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7532],
                                'sighex': salt.utils.stringutils.to_str(other2Data['sighex']),
                                'prihex': salt.utils.stringutils.to_str(other2Data['prihex']),
                                'role': other2Data['role'],
                            })

        # should not join since role same but keys different
        self.join(other2, main)
        self.assertEqual(len(main.transactions), 0)  # rejected since not same keys
        self.assertEqual(len(other2.remotes), 0)
        self.assertEqual(len(main.remotes), 1)
        #main.removeRemote(main.nameRemotes[other2.local.name], clear=True)
        other2.server.close()
        other2.keep.clearAllDir()
        path = os.path.join(main.keep.remotedirpath,
                                "{0}.{1}.{2}".format(main.keep.prefix, other2.local.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))
        other2.keep.clearRoleDir()
        self.assertFalse(os.path.exists(opts['pki_dir']))
        #shutil.rmtree(opts['pki_dir'])

        # recreate other2 stack but use same role and same keys as other1
        opts = self.createOpts(role='primary',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.caller],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        other2Data = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        other2Data['sighex'] = other1Data['sighex']
        other2Data['prihex'] = other1Data['prihex']
        other2Keep = salting.SaltKeep(opts=opts,
                                      basedirpath=other2Data['basedirpath'],
                                      stackname=other2Data['name'])

        self.assertEqual(other2Keep.loadLocalData(), None)
        self.assertEqual(other2Keep.loadAllRemoteData(), {})

        other2 = self.createRoadStack(data=other2Data,
                                     main=None,
                                     ha=("", 7532),
                                     keep=other2Keep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other2.name, other2.keep.dirpath))
        self.assertTrue(other2.keep.dirpath.endswith(os.path.join('primary', 'raet', 'primary_caller')))
        self.assertEqual(other2.ha, ("0.0.0.0", 7532))
        self.assertIs(other2.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other2.keep.loadLocalData(),
                            {
                                'name': other2Data['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7532],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7532],
                                'role': other2Data['role'],
                                'sighex': salt.utils.stringutils.to_str(other1Data['sighex']),
                                'prihex': salt.utils.stringutils.to_str(other1Data['prihex']),
                            })

        # should join since same role and keys
        self.join(other2, main)

        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other2.transactions), 0)
        remote = next(iter(other2.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other2, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other2.transactions), 0)
        remote = next(iter(other2.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # now delete a key and see if both road keep file are also deleted
        main.keep.saltRaetKey.delete_key(match=other1.local.role)
        remote = main.remotes[2]
        path = os.path.join(main.keep.remotedirpath,
                "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))
        remote = main.remotes[4]
        path = os.path.join(main.keep.remotedirpath,
                        "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))

        for stack in [main, other1, other2]:
            stack.server.close()
            stack.clearAllKeeps()

    def testBootstrapRoleAuto(self):
        '''
        Bootstap to allowed with multiple remotes using same role with auto main
        '''
        console.terse("{0}\n".format(self.testBootstrapRoleAuto.__doc__))

        opts = self.createOpts(role='main',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.master],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        mainData = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        mainKeep = salting.SaltKeep(opts=opts,
                                    basedirpath=mainData['basedirpath'],
                                    stackname=mainData['name'])

        self.assertEqual(mainKeep.loadLocalData(), None)
        self.assertEqual(mainKeep.loadAllRemoteData(), {})

        main = self.createRoadStack(data=mainData,
                                     main=True,
                                     ha=None,  # default ha is ("", raeting.RAET_PORT)
                                     keep=mainKeep)

        console.terse("{0}\nkeep dirpath = {1}\n".format(
                main.name, main.keep.dirpath))
        self.assertTrue(main.keep.dirpath.endswith(os.path.join('main', 'raet', 'main_master')))
        self.assertTrue(main.ha, ("0.0.0.0", raeting.RAET_PORT))
        self.assertIs(main.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(main.keep.loadLocalData(), {
                                                         'name': mainData['name'],
                                                         'uid': 1,
                                                         'ha': ['127.0.0.1', 7530],
                                                         'iha': None,
                                                         'natted': None,
                                                         'fqdn': socket.getfqdn('127.0.0.1'),
                                                         'dyned': None,
                                                         'sid': 0,
                                                         'puid': 1,
                                                         'aha': ['0.0.0.0', 7530],
                                                          'role': mainData['role'],
                                                         'sighex': salt.utils.stringutils.to_str(mainData['sighex']),
                                                         'prihex': salt.utils.stringutils.to_str(mainData['prihex']),
                                                         })

        opts = self.createOpts(role='primary',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.minion],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        other1Data = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        other1Keep = salting.SaltKeep(opts=opts,
                                      basedirpath=other1Data['basedirpath'],
                                      stackname=other1Data['name'])

        self.assertEqual(other1Keep.loadLocalData(), None)
        self.assertEqual(other1Keep.loadAllRemoteData(), {})

        other1 = self.createRoadStack(data=other1Data,
                                     main=None,
                                     ha=("", raeting.RAET_TEST_PORT),
                                     keep=other1Keep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other1.name, other1.keep.dirpath))
        self.assertTrue(other1.keep.dirpath.endswith(os.path.join('primary', 'raet', 'primary_minion')))
        self.assertEqual(other1.ha, ("0.0.0.0", raeting.RAET_TEST_PORT))
        self.assertIs(other1.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other1.keep.loadLocalData(),
                            {
                                'name': other1Data['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7531],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7531],
                                'role': other1Data['role'],
                                'sighex': salt.utils.stringutils.to_str(other1Data['sighex']),
                                'prihex': salt.utils.stringutils.to_str(other1Data['prihex']),
                            })

        self.join(other1, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other1.transactions), 0)
        remote = next(iter(other1.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other1, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other1.transactions), 0)
        remote = next(iter(other1.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # create other2 stack but use same role and different keys as other1
        opts = self.createOpts(role='primary',
                               kind=kinds.APPL_KIND_NAMES[kinds.applKinds.caller],
                               dirpath=self.tempDirpath,
                               openMode=False,
                               autoAccept=True)
        other2Data = self.createRoadData(role=opts['id'],
                                       kind=opts['__role'],
                                       cachedirpath=opts['cachedir'])
        other2Data['sighex'] = other1Data['sighex']
        other2Data['prihex'] = other1Data['prihex']

        other2Keep = salting.SaltKeep(opts=opts,
                                      basedirpath=other2Data['basedirpath'],
                                      stackname=other2Data['name'])

        self.assertEqual(other2Keep.loadLocalData(), None)
        self.assertEqual(other2Keep.loadAllRemoteData(), {})

        other2 = self.createRoadStack(data=other2Data,
                                     main=None,
                                     ha=("", 7532),
                                     keep=other2Keep)

        console.terse("{0} keep dirpath = {1}\n".format(
                other2.name, other2.keep.dirpath))
        self.assertTrue(other2.keep.dirpath.endswith(os.path.join('primary', 'raet', 'primary_caller')))
        self.assertEqual(other2.ha, ("0.0.0.0", 7532))
        self.assertIs(other2.keep.auto, raeting.AutoMode.once.value)
        self.assertDictEqual(other2.keep.loadLocalData(),
                            {
                                'name': other2Data['name'],
                                'uid': 1,
                                'ha': ['127.0.0.1', 7532],
                                'iha': None,
                                'natted': None,
                                'fqdn': socket.getfqdn('127.0.0.1'),
                                'dyned': None,
                                'sid': 0,
                                'puid': 1,
                                'aha': ['0.0.0.0', 7532],
                                'role': other2Data['role'],
                                'sighex': salt.utils.stringutils.to_str(other2Data['sighex']),
                                'prihex': salt.utils.stringutils.to_str(other2Data['prihex']),
                            })

        # should join since open mode
        self.join(other2, main)

        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.joined)
        self.assertEqual(len(other2.transactions), 0)
        remote = next(iter(other2.remotes.values()))
        self.assertTrue(remote.joined)

        self.allow(other2, main)
        self.assertEqual(len(main.transactions), 0)
        remote = next(iter(main.remotes.values()))
        self.assertTrue(remote.allowed)
        self.assertEqual(len(other2.transactions), 0)
        remote = next(iter(other2.remotes.values()))
        self.assertTrue(remote.allowed)

        for remote in main.remotes.values():
            path = os.path.join(main.keep.remotedirpath,
                    "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
            self.assertTrue(os.path.exists(path))

        # now delete a key and see if both road keep file are also deleted
        main.keep.saltRaetKey.delete_key(match=other1.local.role)
        remote = main.remotes[2]
        path = os.path.join(main.keep.remotedirpath,
                "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))
        remote = main.remotes[3]
        path = os.path.join(main.keep.remotedirpath,
                        "{0}.{1}.{2}".format(main.keep.prefix, remote.name, main.keep.ext))
        self.assertFalse(os.path.exists(path))

        for stack in [main, other1, other2]:
            stack.server.close()
            stack.clearAllKeeps()


def runOne(test):
    '''
    Unittest Runner
    '''
    test = BasicTestCase(test)
    suite = unittest.TestSuite([test])
    unittest.TextTestRunner(verbosity=2).run(suite)


def runSome():
    '''
    Unittest runner
    '''
    tests = []
    names = ['testBasic',
             'testBasicOpen',
             'testBasicAuto',
             'testBasicRole',
             'testBasicRoleOpen',
             'testBasicRoleAuto',
             'testBootstrapNever',
             'testBootstrapOpen',
             'testBootstrapAuto',
             'testBootstrapRoleNever',
             'testBootstrapRoleAuto',
             ]

    tests.extend(list(map(BasicTestCase, names)))

    suite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=2).run(suite)


def runAll():
    '''
    Unittest runner
    '''
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(BasicTestCase))

    unittest.TextTestRunner(verbosity=2).run(suite)

if __name__ == '__main__' and __package__ is None:

    #console.reinit(verbosity=console.Wordage.concise)

    runAll()  # run all unittests

    #runSome()  # only run some

    #runOne('testBootstrapRoleAuto')



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Alexandru Bleotu (alexandru.bleotu@morganstanley.com)`


    salt.config.schemas.esxcluster
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ESX Cluster configuration schemas
'''

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
from salt.utils.schema import (Schema,
                               DefinitionsSchema,
                               ComplexSchemaItem,
                               DictItem,
                               ArrayItem,
                               IntegerItem,
                               BooleanItem,
                               StringItem,
                               AnyOfItem)


class OptionValueItem(ComplexSchemaItem):
    '''Sechma item of the OptionValue'''

    title = 'OptionValue'
    key = StringItem(title='Key', required=True)
    value = AnyOfItem(items=[StringItem(), BooleanItem(), IntegerItem()])


class AdmissionControlPolicyItem(ComplexSchemaItem):
    '''
    Schema item of the HA admission control policy
    '''

    title = 'Admission Control Policy'

    cpu_failover_percent = IntegerItem(
        title='CPU Failover Percent',
        minimum=0, maximum=100)
    memory_failover_percent = IntegerItem(
        title='Memory Failover Percent',
        minimum=0, maximum=100)


class DefaultVmSettingsItem(ComplexSchemaItem):
    '''
    Schema item of the HA default vm settings
    '''

    title = 'Default VM Settings'

    isolation_response = StringItem(
        title='Isolation Response',
        enum=['clusterIsolationResponse', 'none', 'powerOff', 'shutdown'])
    restart_priority = StringItem(
        title='Restart Priority',
        enum=['clusterRestartPriority', 'disabled', 'high', 'low', 'medium'])


class HAConfigItem(ComplexSchemaItem):
    '''
    Schema item of ESX cluster high availability
    '''

    title = 'HA Configuration'
    description = 'ESX cluster HA configuration json schema item'

    enabled = BooleanItem(
        title='Enabled',
        description='Specifies if HA should be enabled')
    admission_control_enabled = BooleanItem(
        title='Admission Control Enabled')
    admission_control_policy = AdmissionControlPolicyItem()
    default_vm_settings = DefaultVmSettingsItem()
    hb_ds_candidate_policy = StringItem(
        title='Heartbeat Datastore Candidate Policy',
        enum=['allFeasibleDs', 'allFeasibleDsWithUserPreference',
              'userSelectedDs'])
    host_monitoring = StringItem(title='Host Monitoring',
                                 choices=['enabled', 'disabled'])
    options = ArrayItem(min_items=1, items=OptionValueItem())
    vm_monitoring = StringItem(
        title='Vm Monitoring',
        choices=['vmMonitoringDisabled', 'vmAndAppMonitoring',
                 'vmMonitoringOnly'])


class vSANClusterConfigItem(ComplexSchemaItem):
    '''
    Schema item of the ESX cluster vSAN configuration
    '''

    title = 'vSAN Configuration'
    description = 'ESX cluster vSAN configurationi item'

    enabled = BooleanItem(
        title='Enabled',
        description='Specifies if vSAN should be enabled')
    auto_claim_storage = BooleanItem(
        title='Auto Claim Storage',
        description='Specifies whether the storage of member ESXi hosts should '
        'be automatically claimed for vSAN')
    dedup_enabled = BooleanItem(
        title='Enabled',
        description='Specifies dedup should be enabled')
    compression_enabled = BooleanItem(
        title='Enabled',
        description='Specifies if compression should be enabled')


class DRSConfigItem(ComplexSchemaItem):
    '''
    Schema item of the ESX cluster DRS configuration
    '''

    title = 'DRS Configuration'
    description = 'ESX cluster DRS configuration item'

    enabled = BooleanItem(
        title='Enabled',
        description='Specifies if DRS should be enabled')
    vmotion_rate = IntegerItem(
        title='vMotion rate',
        description='Aggressiveness to do automatic vMotions: '
        '1 (least aggressive) - 5 (most aggressive)',
        minimum=1,
        maximum=5)
    default_vm_behavior = StringItem(
        title='Default VM DRS Behavior',
        description='Specifies the default VM DRS behavior',
        enum=['fullyAutomated', 'partiallyAutomated', 'manual'])


class ESXClusterConfigSchema(DefinitionsSchema):
    '''
    Schema of the ESX cluster config
    '''

    title = 'ESX Cluster Configuration Schema'
    description = 'ESX cluster configuration schema'

    ha = HAConfigItem()
    vsan = vSANClusterConfigItem()
    drs = DRSConfigItem()
    vm_swap_placement = StringItem(title='VM Swap Placement')


class ESXClusterEntitySchema(Schema):
    '''Schema of the ESX cluster entity'''

    title = 'ESX Cluster Entity Schema'
    description = 'ESX cluster entity schema'

    type = StringItem(title='Type',
                      description='Specifies the entity type',
                      required=True,
                      enum=['cluster'])

    datacenter = StringItem(title='Datacenter',
                            description='Specifies the cluster datacenter',
                            required=True,
                            pattern=r'\w+')

    cluster = StringItem(title='Cluster',
                         description='Specifies the cluster name',
                         required=True,
                         pattern=r'\w+')


class LicenseSchema(Schema):
    '''
    Schema item of the ESX cluster vSAN configuration
    '''

    title = 'Licenses schema'
    description = 'License configuration schema'

    licenses = DictItem(
        title='Licenses',
        description='Dictionary containing the license name to key mapping',
        required=True,
        additional_properties=StringItem(
            title='License Key',
            description='Specifies the license key',
            pattern=r'^(\w{5}-\w{5}-\w{5}-\w{5}-\w{5})$'))


class EsxclusterProxySchema(Schema):
    '''
    Schema of the esxcluster proxy input
    '''

    title = 'Esxcluster Proxy Schema'
    description = 'Esxcluster proxy schema'
    additional_properties = False
    proxytype = StringItem(required=True,
                           enum=['esxcluster'])
    vcenter = StringItem(required=True, pattern=r'[^\s]+')
    datacenter = StringItem(required=True)
    cluster = StringItem(required=True)
    mechanism = StringItem(required=True, enum=['userpass', 'sspi'])
    username = StringItem()
    passwords = ArrayItem(min_items=1,
                          items=StringItem(),
                          unique_items=True)
    # TODO Should be changed when anyOf is supported for schemas
    domain = StringItem()
    principal = StringItem()
    protocol = StringItem()
    port = IntegerItem(minimum=1)



# -*- coding: utf-8 -*-
'''
An engine that reads messages from Hipchat and sends them to the Salt
event bus.  Alternatively Salt commands can be sent to the Salt master
via Hipchat by setting the control parameter to ``True`` and using command
prefaced with a ``!``. Only token key is required, but room and control
keys make the engine interactive.

.. versionadded: 2016.11.0

:depends: hypchat
:configuration: Example configuration

    .. code-block:: yaml

        engines:
            - hipchat:
                api_url: http://api.hipchat.myteam.com
                token: 'XXXXXX'
                room: 'salt'
                control: True
                valid_users:
                    - SomeUser
                valid_commands:
                    - test.ping
                    - cmd.run
                    - list_jobs
                    - list_commands
                aliases:
                    list_jobs:
                        cmd: jobs.list_jobs
                    list_commands:
                        cmd: pillar.get salt:engines:hipchat:valid_commands target=saltmaster
                max_rooms: 0
                wait_time: 1
'''

from __future__ import absolute_import, print_function, unicode_literals
import logging
import time
import os


try:
    import hypchat
    HAS_HYPCHAT = True
except ImportError:
    HAS_HYPCHAT = False

import salt.utils.args
import salt.utils.event
import salt.utils.files
import salt.utils.http
import salt.utils.json
import salt.utils.stringutils
import salt.runner
import salt.client
import salt.loader
import salt.output
from salt.ext import six


def __virtual__():
    return HAS_HYPCHAT

log = logging.getLogger(__name__)

_DEFAULT_API_URL = 'https://api.hipchat.com'
_DEFAULT_SLEEP = 5
_DEFAULT_MAX_ROOMS = 1000


def _publish_file(token, room, filepath, message='', outputter=None, api_url=None):
    '''
    Send file to a HipChat room via API version 2

    Parameters
    ----------
    token : str
        HipChat API version 2 compatible token - must be token for active user
    room: str
        Name or API ID of the room to notify
    filepath: str
        Full path of file to be sent
    message: str, optional
        Message to send to room
    api_url: str, optional
        Hipchat API URL to use, defaults to http://api.hipchat.com
    '''

    if not os.path.isfile(filepath):
        raise ValueError("File '{0}' does not exist".format(filepath))
    if len(message) > 1000:
        raise ValueError('Message too long')

    url = "{0}/v2/room/{1}/share/file".format(api_url, room)
    headers = {'Content-type': 'multipart/related; boundary=boundary123456'}
    headers['Authorization'] = "Bearer " + token
    msg = salt.utils.json.dumps({'message': message})

    # future lint: disable=blacklisted-function
    with salt.utils.files.fopen(filepath, 'rb') as rfh:
        payload = str('''\
--boundary123456
Content-Type: application/json; charset=UTF-8
Content-Disposition: attachment; name="metadata"

{0}

--boundary123456
Content-Disposition: attachment; name="file"; filename="{1}"

{2}

--boundary123456--\
''').format(msg,
            os.path.basename(salt.utils.stringutils.to_str(filepath)),
            salt.utils.stringutils.to_str(rfh.read()))
    # future lint: enable=blacklisted-function

    salt.utils.http.query(url, method='POST', header_dict=headers, data=payload)


def _publish_html_message(token, room, data, message='', outputter='nested', api_url=None):
    '''
    Publishes the HTML-formatted message.
    '''
    url = "{0}/v2/room/{1}/notification".format(api_url, room)
    headers = {
        'Content-type': 'text/plain'
    }
    headers['Authorization'] = 'Bearer ' + token
    salt.utils.http.query(
        url,
        'POST',
        data=message,
        decode=True,
        status=True,
        header_dict=headers,
        opts=__opts__,
    )
    headers['Content-type'] = 'text/html'
    message = salt.output.html_format(data, outputter, opts=__opts__)
    salt.utils.http.query(
        url,
        'POST',
        data=message,
        decode=True,
        status=True,
        header_dict=headers,
        opts=__opts__,
    )


def _publish_code_message(token, room, data, message='', outputter='nested', api_url=None):
    '''
    Publishes the output format as code.
    '''
    url = "{0}/v2/room/{1}/notification".format(api_url, room)
    headers = {
        'Content-type': 'text/plain'
    }
    headers['Authorization'] = 'Bearer ' + token
    salt.utils.http.query(
        url,
        'POST',
        data=message,
        decode=True,
        status=True,
        header_dict=headers,
        opts=__opts__,
    )
    message = '/code '
    message += salt.output.string_format(data, outputter, opts=__opts__)
    salt.utils.http.query(
        url,
        'POST',
        data=message,
        decode=True,
        status=True,
        header_dict=headers,
        opts=__opts__,
    )


def start(token,
          room='salt',
          aliases=None,
          valid_users=None,
          valid_commands=None,
          control=False,
          trigger="!",
          tag='salt/engines/hipchat/incoming',
          api_key=None,
          api_url=None,
          max_rooms=None,
          wait_time=None,
          output_type='file',
          outputter='nested'):
    '''
    Listen to Hipchat messages and forward them to Salt.

    token
        The HipChat API key. It requires a key for global usgae,
        assigned per user, rather than room.

    room
        The HipChat room name.

    aliases
        Define custom aliases.

    valid_users
        Restrict access only to certain users.

    valid_commands
        Restrict the execution to a limited set of commands.

    control
        Send commands to the master.

    trigger: ``!``
        Special character that triggers the execution of salt commands.

    tag: ``salt/engines/hipchat/incoming``
        The event tag on the Salt bus.

    api_url: ``https://api.hipchat.com``
        The URL to the HipChat API.

        .. versionadded:: 2017.7.0

    max_rooms: ``1000``
        Maximum number of rooms allowed to fetch. If set to 0,
        it is able to retrieve the entire list of rooms.

    wait_time: ``5``
        Maximum wait time, in seconds.

    output_type: ``file``
        The type of the output. Choose bewteen:

            - ``file``: save the output into a temporary file and upload
            - ``html``: send the output as HTML
            - ``code``: send the output as code

        This can be overridden when executing a command, using the ``--out-type`` argument.

        .. versionadded:: 2017.7.0

    outputter: ``nested``
        The format to display the data, using the outputters available on the CLI.
        This argument can also be overridden when executing a command, using the ``--out`` option.

        .. versionadded:: 2017.7.0

    HipChat Example:

    .. code-block:: text

        ! test.ping
        ! test.ping target=minion1
        ! test.ping --out=nested
        ! test.ping --out-type=code --out=table
    '''
    target_room = None

    if __opts__.get('__role') == 'master':
        fire_master = salt.utils.event.get_master_event(
            __opts__,
            __opts__['sock_dir']).fire_event
    else:
        fire_master = None

    def fire(tag, msg):
        '''
        fire event to salt bus
        '''

        if fire_master:
            fire_master(msg, tag)
        else:
            __salt__['event.send'](tag, msg)

    def _eval_bot_mentions(all_messages, trigger):
        ''' yield partner message '''
        for message in all_messages:
            message_text = message['message']
            if message_text.startswith(trigger):
                fire(tag, message)
                text = message_text.replace(trigger, '').strip()
                yield message['from']['mention_name'], text

    token = token or api_key
    if not token:
        raise UserWarning("Hipchat token not found")

    runner_functions = sorted(salt.runner.Runner(__opts__).functions)

    if not api_url:
        api_url = _DEFAULT_API_URL
    hipc = hypchat.HypChat(token, endpoint=api_url)
    if not hipc:
        raise UserWarning("Unable to connect to hipchat")

    log.debug('Connected to Hipchat')
    rooms_kwargs = {}
    if max_rooms is None:
        max_rooms = _DEFAULT_MAX_ROOMS
        rooms_kwargs['max_results'] = max_rooms
    elif max_rooms > 0:
        rooms_kwargs['max_results'] = max_rooms
    # if max_rooms is 0 => retrieve all (rooms_kwargs is empty dict)
    all_rooms = hipc.rooms(**rooms_kwargs)['items']
    for a_room in all_rooms:
        if a_room['name'] == room:
            target_room = a_room
    if not target_room:
        log.debug("Unable to connect to room %s", room)
        # wait for a bit as to not burn through api calls
        time.sleep(30)
        raise UserWarning("Unable to connect to room {0}".format(room))

    after_message_id = target_room.latest(maxResults=1)['items'][0]['id']

    while True:
        try:
            new_messages = target_room.latest(
                not_before=after_message_id)['items']
        except hypchat.requests.HttpServiceUnavailable:
            time.sleep(15)
            continue

        after_message_id = new_messages[-1]['id']
        for partner, text in _eval_bot_mentions(new_messages[1:], trigger):
            # bot summoned by partner

            if not control:
                log.debug("Engine not configured for control")
                return

            # Ensure the user is allowed to run commands
            if valid_users:
                if partner not in valid_users:
                    target_room.message('{0} not authorized to run Salt commands'.format(partner))
                    return

            args = []
            kwargs = {}

            cmdline = salt.utils.args.shlex_split(text)
            cmd = cmdline[0]

            # Evaluate aliases
            if aliases and isinstance(aliases, dict) and cmd in aliases.keys():
                cmdline = aliases[cmd].get('cmd')
                cmdline = salt.utils.args.shlex_split(cmdline)
                cmd = cmdline[0]

            # Parse args and kwargs
            if len(cmdline) > 1:
                for item in cmdline[1:]:
                    if '=' in item:
                        (key, value) = item.split('=', 1)
                        kwargs[key] = value
                    else:
                        args.append(item)

            # Check for target. Otherwise assume *
            if 'target' not in kwargs:
                target = '*'
            else:
                target = kwargs['target']
                del kwargs['target']

            # Check for tgt_type. Otherwise assume glob
            if 'tgt_type' not in kwargs:
                tgt_type = 'glob'
            else:
                tgt_type = kwargs['tgt_type']
                del kwargs['tgt_type']

            # Check for outputter. Otherwise assume nested
            if '--out' in kwargs:
                outputter = kwargs['--out']
                del kwargs['--out']

            # Check for outputter. Otherwise assume nested
            if '--out-type' in kwargs:
                output_type = kwargs['--out-type']
                del kwargs['--out-type']

            # Ensure the command is allowed
            if valid_commands:
                if cmd not in valid_commands:
                    target_room.message('Using {0} is not allowed.'.format(cmd))
                    return

            ret = {}
            if cmd in runner_functions:
                runner = salt.runner.RunnerClient(__opts__)
                ret = runner.cmd(cmd, arg=args, kwarg=kwargs)

            # Default to trying to run as a client module.
            else:
                local = salt.client.LocalClient()
                ret = local.cmd('{0}'.format(target), cmd, args, kwargs, tgt_type='{0}'.format(tgt_type))

            nice_args = (' ' + ' '.join(args)) if args else ''
            nice_kwargs = (' ' + ' '.join('{0}={1}'.format(key, value) for (key, value) in six.iteritems(kwargs))) \
                          if kwargs else ''
            message_string = '@{0} Results for: {1}{2}{3} on {4}'.format(partner,
                                                                         cmd,
                                                                         nice_args,
                                                                         nice_kwargs,
                                                                         target)
            if output_type == 'html':
                _publish_html_message(token, room, ret, message=message_string, outputter=outputter, api_url=api_url)
            elif output_type == 'code':
                _publish_code_message(token, room, ret, message=message_string, outputter=outputter, api_url=api_url)
            else:
                tmp_path_fn = salt.utils.files.mkstemp()
                with salt.utils.files.fopen(tmp_path_fn, 'w+') as fp_:
                    salt.utils.json.dump(ret, fp_, sort_keys=True, indent=4)
                _publish_file(token, room, tmp_path_fn, message=message_string, api_url=api_url)
                salt.utils.files.safe_rm(tmp_path_fn)
        time.sleep(wait_time or _DEFAULT_SLEEP)



# -*- coding: utf-8 -*-
'''
Kapacitor execution module.

:configuration: This module accepts connection configuration details either as
    parameters or as configuration settings in /etc/salt/minion on the relevant
    minions::

        kapacitor.host: 'localhost'
        kapacitor.port: 9092

    This data can also be passed into pillar. Options passed into opts will
    overwrite options passed into pillar.

.. versionadded:: 2016.11.0
'''
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt lobs
from salt.ext import six
import salt.utils.http
import salt.utils.json
import salt.utils.path
from salt.utils.decorators import memoize


def __virtual__():
    return 'kapacitor' if salt.utils.path.which('kapacitor') else False


@memoize
def version():
    '''
    Get the kapacitor version.
    '''
    version = __salt__['pkg.version']('kapacitor')
    if not version:
        version = six.string_types(__salt__['config.option']('kapacitor.version', 'latest'))
    return version


def get_task(name):
    '''
    Get a dict of data on a task.

    name
        Name of the task to get information about.

    CLI Example:

    .. code-block:: bash

        salt '*' kapacitor.get_task cpu
    '''
    host = __salt__['config.option']('kapacitor.host', 'localhost')
    port = __salt__['config.option']('kapacitor.port', 9092)

    if version() < '0.13':
        url = 'http://{0}:{1}/task?name={2}'.format(host, port, name)
    else:
        url = 'http://{0}:{1}/kapacitor/v1/tasks/{2}?skip-format=true'.format(host, port, name)

    response = salt.utils.http.query(url, status=True)

    if response['status'] == 404:
        return None

    data = salt.utils.json.loads(response['body'])

    if version() < '0.13':
        return {
            'script': data['TICKscript'],
            'type': data['Type'],
            'dbrps': data['DBRPs'],
            'enabled': data['Enabled'],
        }

    return {
        'script': data['script'],
        'type': data['type'],
        'dbrps': data['dbrps'],
        'enabled': data['status'] == 'enabled',
    }


def _run_cmd(cmd):
    '''
    Run a Kapacitor task and return a dictionary of info.
    '''
    ret = {}
    result = __salt__['cmd.run_all'](cmd)

    if result.get('stdout'):
        ret['stdout'] = result['stdout']
    if result.get('stderr'):
        ret['stderr'] = result['stderr']
    ret['success'] = result['retcode'] == 0

    return ret


def define_task(name,
                tick_script,
                task_type='stream',
                database=None,
                retention_policy='default'):
    '''
    Define a task. Serves as both create/update.

    name
        Name of the task.

    tick_script
        Path to the TICK script for the task. Can be a salt:// source.

    task_type
        Task type. Defaults to 'stream'

    database
        Which database to fetch data from. Defaults to None, which will use the
        default database in InfluxDB.

    retention_policy
        Which retention policy to fetch data from. Defaults to 'default'.

    CLI Example:

    .. code-block:: bash

        salt '*' kapacitor.define_task cpu salt://kapacitor/cpu.tick database=telegraf
    '''
    if version() < '0.13':
        cmd = 'kapacitor define -name {0}'.format(name)
    else:
        cmd = 'kapacitor define {0}'.format(name)

    if tick_script.startswith('salt://'):
        tick_script = __salt__['cp.cache_file'](tick_script, __env__)

    cmd += ' -tick {0}'.format(tick_script)

    if task_type:
        cmd += ' -type {0}'.format(task_type)

    if database and retention_policy:
        cmd += ' -dbrp {0}.{1}'.format(database, retention_policy)

    return _run_cmd(cmd)


def delete_task(name):
    '''
    Delete a kapacitor task.

    name
        Name of the task to delete.

    CLI Example:

    .. code-block:: bash

        salt '*' kapacitor.delete_task cpu
    '''
    return _run_cmd('kapacitor delete tasks {0}'.format(name))


def enable_task(name):
    '''
    Enable a kapacitor task.

    name
        Name of the task to enable.

    CLI Example:

    .. code-block:: bash

        salt '*' kapacitor.enable_task cpu
    '''
    return _run_cmd('kapacitor enable {0}'.format(name))


def disable_task(name):
    '''
    Disable a kapacitor task.

    name
        Name of the task to disable.

    CLI Example:

    .. code-block:: bash

        salt '*' kapacitor.disable_task cpu
    '''
    return _run_cmd('kapacitor disable {0}'.format(name))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, unicode_literals, print_function

# Import Python libs
import logging
log = logging.getLogger(__name__)

# Import pyrax (SDK for Rackspace cloud) third party libs
try:
    import pyrax
    import pyrax.exceptions
except ImportError:
    raise

# Import salt classes
from salt.utils.openstack.pyrax import authenticate


class RackspaceQueues(object):
    def __init__(self, username, password, region, **kwargs):
        self.auth = authenticate.Authenticate(username, password, region, **kwargs)
        self.conn = self.auth.conn.queues

    def create(self, qname):
        '''
        Create RackSpace Queue.
        '''
        try:
            if self.exists(qname):
                log.error('Queues "%s" already exists. Nothing done.', qname)
                return True

            self.conn.create(qname)

            return True
        except pyrax.exceptions as err_msg:
            log.error('RackSpace API got some problems during creation: %s',
                      err_msg)
        return False

    def delete(self, qname):
        '''
        Delete an existings RackSpace Queue.
        '''
        try:
            q = self.exists(qname)
            if not q:
                return False
            queue = self.show(qname)
            if queue:
                queue.delete()
        except pyrax.exceptions as err_msg:
            log.error('RackSpace API got some problems during deletion: %s',
                      err_msg)
            return False

        return True

    def exists(self, qname):
        '''
        Check to see if a Queue exists.
        '''
        try:
            # First if not exists() -> exit
            if self.conn.queue_exists(qname):
                return True
            return False
        except pyrax.exceptions as err_msg:
            log.error('RackSpace API got some problems during '
                      'existing queue check: %s',
                      err_msg)
        return False

    def show(self, qname):
        '''
        Show information about Queue
        '''
        try:
            # First if not exists() -> exit
            if not self.conn.queue_exists(qname):
                return {}
            # If exist, search the queue to return the Queue Object
            for queue in self.conn.list():
                if queue.name == qname:
                    return queue
        except pyrax.exceptions as err_msg:
            log.error('RackSpace API got some problems during existing'
                      ' queue check: %s', err_msg)
        return {}



# -*- coding: utf-8 -*-
'''
Monitor Web Server with Uptime
==============================

`Uptime <https://github.com/fzaninotto/uptime>`_ is an open source
remote monitoring application using Node.js, MongoDB, and Twitter
Bootstrap.

.. warning::

    This state module is beta. It might be changed later to include
    more or less automation.

.. note::

    This state module requires a pillar to specify the location of
    your uptime install

    .. code-block:: yaml

        uptime:
          application_url: "http://uptime-url.example.org"

Example:

.. code-block:: yaml

    url:
      uptime.monitored
    url/sitemap.xml:
      uptime.monitored:
         - polling: 600 # every hour

'''
# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals


def __virtual__():
    '''
    Only load if the uptime module is present
    '''
    return 'uptime.checks_list' in __salt__


def monitored(name, **params):
    '''
    Makes sure an URL is monitored by uptime. Checks if URL is already
    monitored, and if not, adds it.

    '''

    ret = {'name': name, 'changes': {}, 'result': None, 'comment': ''}
    if __salt__['uptime.check_exists'](name=name):
        ret['result'] = True
        ret['comment'] = 'URL {0} is already monitored'.format(name)
        ret['changes'] = {}
        return ret
    if not __opts__['test']:
        url_monitored = __salt__['uptime.create'](name, **params)
        if url_monitored:
            ret['result'] = True
            msg = 'Successfully added the URL {0} to uptime'
            ret['comment'] = msg.format(name)
            ret['changes'] = {'url_monitored': url_monitored}
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to add {0} to uptime'.format(name)
            ret['changes'] = {}
    else:
        msg = 'URL {0} is going to be added to uptime'
        ret.update(result=None,
                   comment=msg.format(name))
    return ret



# -*- coding: utf-8 -*-
'''
Generate baseline proxy minion grains
'''
from __future__ import absolute_import, print_function, unicode_literals
import salt.utils.platform

__proxyenabled__ = ['ssh_sample']

__virtualname__ = 'ssh_sample'


def __virtual__():
    try:
        if salt.utils.platform.is_proxy() and __opts__['proxy']['proxytype'] == 'ssh_sample':
            return __virtualname__
    except KeyError:
        pass

    return False


def kernel():
    return {'kernel': 'proxy'}


def proxy_functions(proxy):
    '''
    The loader will execute functions with one argument and pass
    a reference to the proxymodules LazyLoader object.  However,
    grains sometimes get called before the LazyLoader object is setup
    so `proxy` might be None.
    '''
    return {'proxy_functions': proxy['ssh_sample.fns']()}


def location():
    return {'location': 'At the other end of an SSH Tunnel!!'}


def os_data():
    return {'os_data': 'DumbShell Endpoint release 4.09.g'}



# -*- coding: utf-8 -*-
'''
Support for Config Server Firewall (CSF)
========================================
:maintainer: Mostafa Hussein <mostafa.hussein91@gmail.com>
:maturity: new
:platform: Linux
'''

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals
import re

# Import Salt Libs
import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.ext import six


def __virtual__():
    '''
    Only load if csf exists on the system
    '''
    if salt.utils.path.which('csf') is None:
        return (False,
                'The csf execution module cannot be loaded: csf unavailable.')
    else:
        return True


def _temp_exists(method, ip):
    '''
    Checks if the ip exists as a temporary rule based
    on the method supplied, (tempallow, tempdeny).
    '''
    _type = method.replace('temp', '').upper()
    cmd = "csf -t | awk -v code=1 -v type=_type -v ip=ip '$1==type && $2==ip {{code=0}} END {{exit code}}'".format(_type=_type, ip=ip)
    exists = __salt__['cmd.run_all'](cmd)
    return not bool(exists['retcode'])


def _exists_with_port(method, rule):
    path = '/etc/csf/csf.{0}'.format(method)
    return __salt__['file.contains'](path, rule)


def exists(method,
            ip,
            port=None,
            proto='tcp',
            direction='in',
            port_origin='d',
            ip_origin='d',
            ttl=None,
            comment=''):
    '''
    Returns true a rule for the ip already exists
    based on the method supplied. Returns false if
    not found.
    CLI Example:

    .. code-block:: bash

        salt '*' csf.exists allow 1.2.3.4
        salt '*' csf.exists tempdeny 1.2.3.4
    '''
    if method.startswith('temp'):
        return _temp_exists(method, ip)
    if port:
        rule = _build_port_rule(ip, port, proto, direction, port_origin, ip_origin, comment)
        return _exists_with_port(method, rule)
    exists = __salt__['cmd.run_all']("egrep ^'{0} +' /etc/csf/csf.{1}".format(ip, method))
    return not bool(exists['retcode'])


def __csf_cmd(cmd):
    '''
    Execute csf command
    '''
    csf_cmd = '{0} {1}'.format(salt.utils.path.which('csf'), cmd)
    out = __salt__['cmd.run_all'](csf_cmd)

    if out['retcode'] != 0:
        if not out['stderr']:
            ret = out['stdout']
        else:
            ret = out['stderr']
        raise CommandExecutionError(
            'csf failed: {0}'.format(ret)
        )
    else:
        ret = out['stdout']
    return ret


def _status_csf():
    '''
    Return True if csf is running otherwise return False
    '''
    cmd = 'test -e /etc/csf/csf.disable'
    out = __salt__['cmd.run_all'](cmd)
    return bool(out['retcode'])


def _get_opt(method):
    '''
    Returns the cmd option based on a long form argument.
    '''
    opts = {
        'allow': '-a',
        'deny': '-d',
        'unallow': '-ar',
        'undeny': '-dr',
        'tempallow': '-ta',
        'tempdeny': '-td',
        'temprm': '-tr'
    }
    return opts[method]


def _build_args(method, ip, comment):
    '''
    Returns the cmd args for csf basic allow/deny commands.
    '''
    opt = _get_opt(method)
    args = '{0} {1}'.format(opt, ip)
    if comment:
        args += ' {0}'.format(comment)
    return args


def _access_rule(method,
                ip=None,
                port=None,
                proto='tcp',
                direction='in',
                port_origin='d',
                ip_origin='d',
                comment=''):
    '''
    Handles the cmd execution for allow and deny commands.
    '''
    if _status_csf():
        if ip is None:
            return {'error': 'You must supply an ip address or CIDR.'}
        if port is None:
            args = _build_args(method, ip, comment)
            return __csf_cmd(args)
        else:
            if method not in ['allow', 'deny']:
                return {'error': 'Only allow and deny rules are allowed when specifying a port.'}
            return _access_rule_with_port(method=method,
                                            ip=ip,
                                            port=port,
                                            proto=proto,
                                            direction=direction,
                                            port_origin=port_origin,
                                            ip_origin=ip_origin,
                                            comment=comment)


def _build_port_rule(ip, port, proto, direction, port_origin, ip_origin, comment):
    kwargs = {
        'ip': ip,
        'port': port,
        'proto': proto,
        'direction': direction,
        'port_origin': port_origin,
        'ip_origin': ip_origin,
    }
    rule = '{proto}|{direction}|{port_origin}={port}|{ip_origin}={ip}'.format(**kwargs)
    if comment:
        rule += ' #{0}'.format(comment)

    return rule


def _remove_access_rule_with_port(method,
                                    ip,
                                    port,
                                    proto='tcp',
                                    direction='in',
                                    port_origin='d',
                                    ip_origin='d',
                                    ttl=None):

    rule = _build_port_rule(ip,
                            port=port,
                            proto=proto,
                            direction=direction,
                            port_origin=port_origin,
                            ip_origin=ip_origin,
                            comment='')

    rule = rule.replace('|', '[|]')
    rule = rule.replace('.', '[.]')
    result = __salt__['file.replace']('/etc/csf/csf.{0}'.format(method),
            pattern='^{0}(( +)?\#.*)?$\n'.format(rule),  # pylint: disable=W1401
                                        repl='')

    return result


def _csf_to_list(option):
    '''
    Extract comma-separated values from a csf.conf
    option and return a list.
    '''
    result = []
    line = get_option(option)
    if line:
        csv = line.split('=')[1].replace(' ', '').replace('"', '')
        result = csv.split(',')
    return result


def split_option(option):
    l = re.split("(?: +)?\=(?: +)?", option)  # pylint: disable=W1401
    return l


def get_option(option):
    pattern = '^{0}(\ +)?\=(\ +)?".*"$'.format(option)  # pylint: disable=W1401
    grep = __salt__['file.grep']('/etc/csf/csf.conf', pattern, '-E')
    if 'stdout' in grep and grep['stdout']:
        line = grep['stdout']
        return line
    return None


def set_option(option, value):
    current_option = get_option(option)
    if not current_option:
        return {'error': 'No such option exists in csf.conf'}
    result = __salt__['file.replace']('/etc/csf/csf.conf',
            pattern='^{0}(\ +)?\=(\ +)?".*"'.format(option),  # pylint: disable=W1401
                                        repl='{0} = "{1}"'.format(option, value))

    return result


def get_skipped_nics(ipv6=False):
    if ipv6:
        option = 'ETH6_DEVICE_SKIP'
    else:
        option = 'ETH_DEVICE_SKIP'

    skipped_nics = _csf_to_list(option)
    return skipped_nics


def skip_nic(nic, ipv6=False):
    nics = get_skipped_nics(ipv6=ipv6)
    nics.append(nic)
    return skip_nics(nics, ipv6)


def skip_nics(nics, ipv6=False):
    if ipv6:
        ipv6 = '6'
    else:
        ipv6 = ''
    nics_csv = ','.join(six.moves.map(six.text_type, nics))
    result = __salt__['file.replace']('/etc/csf/csf.conf',
            pattern='^ETH{0}_DEVICE_SKIP(\ +)?\=(\ +)?".*"'.format(ipv6),  # pylint: disable=W1401
                                        repl='ETH{0}_DEVICE_SKIP = "{1}"'.format(ipv6, nics_csv))

    return result


def _access_rule_with_port(method,
                            ip,
                            port,
                            proto='tcp',
                            direction='in',
                            port_origin='d',
                            ip_origin='d',
                            ttl=None,
                            comment=''):

    results = {}
    if direction == 'both':
        directions = ['in', 'out']
    else:
        directions = [direction]
    for direction in directions:
        _exists = exists(method,
                            ip,
                            port=port,
                            proto=proto,
                            direction=direction,
                            port_origin=port_origin,
                            ip_origin=ip_origin,
                            ttl=ttl,
                            comment=comment)
        if not _exists:
            rule = _build_port_rule(ip,
                                    port=port,
                                    proto=proto,
                                    direction=direction,
                                    port_origin=port_origin,
                                    ip_origin=ip_origin,
                                    comment=comment)
            path = '/etc/csf/csf.{0}'.format(method)
            results[direction] = __salt__['file.append'](path, rule)
    return results


def _tmp_access_rule(method,
                    ip=None,
                    ttl=None,
                    port=None,
                    direction='in',
                    port_origin='d',
                    ip_origin='d',
                    comment=''):
    '''
    Handles the cmd execution for tempdeny and tempallow commands.
    '''
    if _status_csf():
        if ip is None:
            return {'error': 'You must supply an ip address or CIDR.'}
        if ttl is None:
            return {'error': 'You must supply a ttl.'}
        args = _build_tmp_access_args(method, ip, ttl, port, direction, comment)
        return __csf_cmd(args)


def _build_tmp_access_args(method, ip, ttl, port, direction, comment):
    '''
    Builds the cmd args for temporary access/deny opts.
    '''
    opt = _get_opt(method)
    args = '{0} {1} {2}'.format(opt, ip, ttl)
    if port:
        args += ' -p {0}'.format(port)
    if direction:
        args += ' -d {0}'.format(direction)
    if comment:
        args += ' #{0}'.format(comment)
    return args


def running():
    '''
    Check csf status
    CLI Example:

    .. code-block:: bash

        salt '*' csf.running
    '''
    return _status_csf()


def disable():
    '''
    Disable csf permanently
    CLI Example:

    .. code-block:: bash

        salt '*' csf.disable
    '''
    if _status_csf():
        return __csf_cmd('-x')


def enable():
    '''
    Activate csf if not running
    CLI Example:

    .. code-block:: bash

        salt '*' csf.enable
    '''
    if not _status_csf():
        return __csf_cmd('-e')


def reload():
    '''
    Restart csf
    CLI Example:

    .. code-block:: bash

        salt '*' csf.reload
    '''
    return __csf_cmd('-r')


def tempallow(ip=None, ttl=None, port=None, direction=None, comment=''):
    '''
    Add an rule to the temporary ip allow list.
    See :func:`_access_rule`.
    1- Add an IP:
    CLI Example:

    .. code-block:: bash

        salt '*' csf.tempallow 127.0.0.1 3600 port=22 direction='in' comment='# Temp dev ssh access'
    '''
    return _tmp_access_rule('tempallow', ip, ttl, port, direction, comment)


def tempdeny(ip=None, ttl=None, port=None, direction=None, comment=''):
    '''
    Add a rule to the temporary ip deny list.
    See :func:`_access_rule`.
    1- Add an IP:
    CLI Example:

    .. code-block:: bash

        salt '*' csf.tempdeny 127.0.0.1 300 port=22 direction='in' comment='# Brute force attempt'
    '''
    return _tmp_access_rule('tempdeny', ip, ttl, port, direction, comment)


def allow(ip,
        port=None,
        proto='tcp',
        direction='in',
        port_origin='d',
        ip_origin='s',
        ttl=None,
        comment=''):
    '''
    Add an rule to csf allowed hosts
    See :func:`_access_rule`.
    1- Add an IP:
    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow 127.0.0.1
        salt '*' csf.allow 127.0.0.1 comment="Allow localhost"
    '''
    return _access_rule('allow',
                        ip,
                        port=port,
                        proto=proto,
                        direction=direction,
                        port_origin=port_origin,
                        ip_origin=ip_origin,
                        comment=comment)


def deny(ip,
        port=None,
        proto='tcp',
        direction='in',
        port_origin='d',
        ip_origin='d',
        ttl=None,
        comment=''):
    '''
    Add an rule to csf denied hosts
    See :func:`_access_rule`.
    1- Deny an IP:
    CLI Example:

    .. code-block:: bash

        salt '*' csf.deny 127.0.0.1
        salt '*' csf.deny 127.0.0.1 comment="Too localhosty"
    '''
    return _access_rule('deny', ip, port, proto, direction, port_origin, ip_origin, comment)


def remove_temp_rule(ip):
    opt = _get_opt('temprm')
    args = '{0} {1}'.format(opt, ip)
    return __csf_cmd(args)


def unallow(ip):
    '''
    Remove a rule from the csf denied hosts
    See :func:`_access_rule`.
    1- Deny an IP:
    CLI Example:

    .. code-block:: bash

        salt '*' csf.unallow 127.0.0.1
    '''
    return _access_rule('unallow', ip)


def undeny(ip):
    '''
    Remove a rule from the csf denied hosts
    See :func:`_access_rule`.
    1- Deny an IP:
    CLI Example:

    .. code-block:: bash

        salt '*' csf.undeny 127.0.0.1
    '''
    return _access_rule('undeny', ip)


def remove_rule(method,
                ip,
                port=None,
                proto='tcp',
                direction='in',
                port_origin='d',
                ip_origin='s',
                ttl=None,
                comment=''):

    if method.startswith('temp') or ttl:
        return remove_temp_rule(ip)

    if not port:
        if method == 'allow':
            return unallow(ip)
        elif method == 'deny':
            return undeny(ip)

    if port:
        return _remove_access_rule_with_port(method=method,
                                            ip=ip,
                                            port=port,
                                            proto=proto,
                                            direction=direction,
                                            port_origin=port_origin,
                                            ip_origin=ip_origin)


def allow_ports(ports, proto='tcp', direction='in'):
    '''
    Fully replace the incoming or outgoing ports
    line in the csf.conf file - e.g. TCP_IN, TCP_OUT,
    UDP_IN, UDP_OUT, etc.

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow_ports ports="[22,80,443,4505,4506]" proto='tcp' direction='in'
    '''

    results = []
    ports = set(ports)
    ports = list(ports)
    proto = proto.upper()
    direction = direction.upper()
    _validate_direction_and_proto(direction, proto)
    ports_csv = ','.join(six.moves.map(six.text_type, ports))
    directions = build_directions(direction)

    for direction in directions:
        result = __salt__['file.replace']('/etc/csf/csf.conf',
                pattern='^{0}_{1}(\ +)?\=(\ +)?".*"$'.format(proto, direction),  # pylint: disable=W1401
                                    repl='{0}_{1} = "{2}"'.format(proto, direction, ports_csv))
        results.append(result)

    return results


def get_ports(proto='tcp', direction='in'):
    '''
    Lists ports from csf.conf based on direction and protocol.
    e.g. - TCP_IN, TCP_OUT, UDP_IN, UDP_OUT, etc..

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow_port 22 proto='tcp' direction='in'
    '''

    proto = proto.upper()
    direction = direction.upper()
    results = {}
    _validate_direction_and_proto(direction, proto)
    directions = build_directions(direction)
    for direction in directions:
        option = '{0}_{1}'.format(proto, direction)
        results[direction] = _csf_to_list(option)

    return results


def _validate_direction_and_proto(direction, proto):
    if direction.upper() not in ['IN', 'OUT', 'BOTH']:
        raise SaltInvocationError(
            'You must supply a direction of in, out, or both'
        )
    if proto.upper() not in ['TCP', 'UDP', 'TCP6', 'UDP6']:
        raise SaltInvocationError(
            'You must supply tcp, udp, tcp6, or udp6 for the proto keyword'
        )
    return


def build_directions(direction):
    direction = direction.upper()
    if direction == 'BOTH':
        directions = ['IN', 'OUT']
    else:
        directions = [direction]
    return directions


def allow_port(port, proto='tcp', direction='both'):
    '''
    Like allow_ports, but it will append to the
    existing entry instead of replacing it.
    Takes a single port instead of a list of ports.

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow_port 22 proto='tcp' direction='in'
    '''

    ports = get_ports(proto=proto, direction=direction)
    direction = direction.upper()
    _validate_direction_and_proto(direction, proto)
    directions = build_directions(direction)
    results = []
    for direction in directions:
        _ports = ports[direction]
        _ports.append(port)
        results += allow_ports(_ports, proto=proto, direction=direction)
    return results


def get_testing_status():
    testing = _csf_to_list('TESTING')[0]
    return testing


def _toggle_testing(val):
    if val == 'on':
        val = '1'
    elif val == 'off':
        val = '0'
    else:
        raise SaltInvocationError(
            "Only valid arg is 'on' or 'off' here."
        )

    result = __salt__['file.replace']('/etc/csf/csf.conf',
            pattern='^TESTING(\ +)?\=(\ +)?".*"',  # pylint: disable=W1401
                                        repl='TESTING = "{0}"'.format(val))
    return result


def enable_testing_mode():
    return _toggle_testing('on')


def disable_testing_mode():
    return _toggle_testing('off')
